/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "auth.h"
#include "cookie.h"

#include <stdlib.h>

typedef struct _AuthModuleList AuthModuleList;

struct _AuthModuleList {
	AuthModuleList *next;

	AuthModule module;
};

AuthMethod auth_methods;

static AuthModuleList *auth_modules;
static AuthReplyData failure_reply;

void auth_register_module(AuthModule *module)
{
	AuthModuleList *list;

	i_assert((auth_methods & module->method) == 0);

	auth_methods |= module->method;

	list = i_new(AuthModuleList, 1);
	memcpy(&list->module, module, sizeof(AuthModule));

	list->next = auth_modules;
	auth_modules = list;
}

void auth_unregister_module(AuthModule *module)
{
	AuthModuleList **pos, *list;

	i_assert(auth_methods & module->method);

        auth_methods &= ~module->method;

	for (pos = &auth_modules; *pos != NULL; pos = &(*pos)->next) {
		if ((*pos)->module.method == module->method) {
			list = *pos;
			*pos = (*pos)->next;
			i_free(list);
			break;
		}
	}
}

void auth_init_request(AuthInitRequestData *request,
		       AuthCallback callback, void *user_data)
{
	AuthModuleList *list;

	if ((auth_methods & request->method) == 0) {
		/* unsupported method */
		i_error("BUG: imap-login requested unsupported "
			"auth method %d", request->method);
		failure_reply.id = request->id;
		callback(&failure_reply, NULL, user_data);
		return;
	}

	for (list = auth_modules; list != NULL; list = list->next) {
		if (list->module.method == request->method) {
			list->module.init(request, callback, user_data);
			return;
		}
	}

	i_assert(0);
}

void auth_continue_request(AuthContinuedRequestData *request,
			   const unsigned char *data,
			   AuthCallback callback, void *user_data)
{
	CookieData *cookie_data;

	cookie_data = cookie_lookup(request->cookie);
	if (cookie_data == NULL) {
		/* timeouted cookie */
		failure_reply.id = request->id;
		callback(&failure_reply, NULL, user_data);
	} else {
		cookie_data->auth_continue(cookie_data, request, data,
					   callback, user_data);
	}
}

extern AuthModule auth_plain;

void auth_init(void)
{
	char *const *methods;
	const char *env;

        auth_modules = NULL;
	auth_methods = 0;

	memset(&failure_reply, 0, sizeof(failure_reply));
	failure_reply.result = AUTH_RESULT_FAILURE;

	/* register wanted methods */
	env = getenv("METHODS");
	if (env == NULL || *env == '\0')
		i_fatal("METHODS environment is unset");

	methods = t_strsplit(env, " ");
	while (*methods != NULL) {
		if (strcasecmp(*methods, "plain") == 0)
			auth_register_module(&auth_plain);
		else {
			i_fatal("Unknown authentication method '%s'",
				*methods);
		}
		methods++;
	}
}

void auth_deinit(void)
{
	auth_unregister_module(&auth_plain);
}
